from . import classification
from . import landsat_functions
from . import river_analysis
from . import sample
from . import sentinel_functions
from . import shoreline_analysis
from . import terrain_analysis
from . import time_series_analysis
from . import zonal_statistics



"""Reduce images in a region of interest"""
from typing import Sequence, Union
import ee
import pandas as pd


ee.Initialize()


def zonal_statistics_image(
    image: ee.Image,
    geom: ee.Geometry,
    scale: float,
    bands: Union[Sequence[str], str] = "all",
    reducer: Union[ee.Reducer, str] = "all",
    tile_scale: int = 16,
) -> pd.DataFrame:
    """
    Function to calculate a statistic in the specified region for one image.

    Parameters
    ----------
    image : ee.Image
        Image of interest.

    geom : ee.Geometry
        Region of interest to reduce the image.

    scale : float
        Pixel size for the sample to perform the zonal statistics.

    bands : Sequence | str (optional)
        List, tuple with the bands of interest or, if you only want one band, the name
        of the band. By default the process takes into consideration all bands.

    reducer : ee.Reducer | str (optional)
        Reducer to apply to the image. By default, image are reduced to its, mean, standard
        deviation, maximum, minimum, and count.

    tile_scale : int (optional)
        Scale of the mosaic to allow EarthEngine to split the task to more cores.

    Returns
    -------
    data : pandas.DataFrame
        DataFrame with all the stats for all spcified bands.
    """
    if bands != "all":
        image = image.select(bands)

    if reducer == "all":
        reducer = ee.Reducer.mean().combine(
            ee.Reducer.stdDev().combine(
                ee.Reducer.max().combine(
                    ee.Reducer.min().combine(ee.Reducer.count(), sharedInputs=True),
                    sharedInputs=True,
                ),
                sharedInputs=True,
            ),
            sharedInputs=True,
        )

    stats = image.reduceRegion(
        reducer=reducer, geometry=geom, scale=scale, tileScale=tile_scale
    )
    stats = stats.set("system:time_start", image.get("system:time_start"))

    data = pd.DataFrame(stats.getInfo(), index=[0])
    data["date"] = pd.DatetimeIndex(
        pd.to_datetime(data["system:time_start"], unit="ms").dt.date
    )

    return data


def zonal_statistics_collection(
    image_collection: ee.ImageCollection,
    geom: ee.Geometry,
    scale: float,
    bands: Union[Sequence[str], str] = "all",
    reducer: Union[ee.Reducer, str] = "all",
    tile_scale: int = 16,
) -> pd.DataFrame:
    """
    Function to calculate a statistic in the specified region for all Image in an image
    collection.

    Parameters
    ----------
    ImageCollection : ee.ImageCollection
        Image Collection with the image to reduce.

    geom : ee.Geometry
        Region of interest to reduce the images.

    scale : float
        Pixel size for the sample to perform the zonal statistics.

    bands : Sequence | str (optional)
        List, tuple with the bands of interest or, if you only want one band, the name
        of the band. By default the process takes into consideration all bands.

    reducer : ee.Reducer | str (optional)
        Reducer to apply to all image. By default, image are reduced to its, mean, standard
        deviation, maximum, minimum, and count.

    tile_scale : int (optional)
        Scale of the mosaic to allow EarthEngine to split the task to more cores.

    Returns
    ------
    data : pandas.DataFrame
        DataFrame with all the stats for all spcified bands.
    """
    if bands != "all":
        image_collection = image_collection.select(bands)

    if reducer == "all":
        reducer = ee.Reducer.mean().combine(
            ee.Reducer.stdDev().combine(
                ee.Reducer.max().combine(
                    ee.Reducer.min().combine(ee.Reducer.count(), sharedInputs=True),
                    sharedInputs=True,
                ),
                sharedInputs=True,
            ),
            sharedInputs=True,
        )

    def reduce_image(image):
        stats = image.reduceRegion(
            reducer=reducer, geometry=geom, scale=scale, tileScale=tile_scale
        )

        stats = ee.Feature(geom, stats)
        stats = stats.set("system:time_start", image.get("system:time_start"))

        return stats

    fc = ee.FeatureCollection(image_collection.map(reduce_image))

    entries = [feature["properties"] for feature in fc.getInfo()["features"]]

    data = pd.DataFrame(entries)
    data["date"] = pd.DatetimeIndex(
        pd.to_datetime(data["system:time_start"], unit="ms").dt.date
    )

    return data



"""Analyse rivers with sattelite images"""
import ee

ee.Initialize()


def plume_characterization(
    image: ee.Image,
    sample_region: ee.Geometry,
    blue: str,
    green: str,
    red: str,
    nir: str,
) -> ee.Image:
    """
    Extract river plume from an image based on the colors in a sample region.

    Parameters
    ----------
    Image : ee.Image
        Image to be classified.

    sample_region : ee.Geometry
        Polygon that enclosess a region of the Image that the user indetifies as
        river plume.

    blue : str
        Blue band name.

    green : str
        Green band name.

    red : str
        Red band name.

    nir : str
        Near infrared band name.

    Returns
    -------
    image : ee.Image
        Plume characterized in the three color bands.
    """
    image = image.addBands(
        image.expression(
            f"(b('{green}') - b('{nir}'))/(b('{green}') + b('{nir}'))"
        ).rename("NDWI"),
        None,
        True,
    )

    mask = image.select("NDWI").gt(0)
    image = image.updateMask(mask)

    reducer = ee.Reducer.min().combine(reducer2=ee.Reducer.max(), sharedInputs=True)

    limits = image.select([blue, green, red]).reduceRegion(
        reducer=reducer, geometry=sample_region, scale=30
    )

    image = image.addBands(
        image.expression(
            f"(b('{blue}') > MINI && b('{blue}') < MAXI) ? 1 : 0",
            {
                "MAXI": limits.getNumber(blue + "_max"),
                "MINI": limits.getNumber(blue + "_min"),
            },
        ).rename("plume_blue"),
        None,
        True,
    )

    image = image.addBands(
        image.expression(
            f"(b('{green}') > MINI && b('{green}') < MAXI) ? 1 : 0",
            {
                "MAXI": limits.getNumber(green + "_max"),
                "MINI": limits.getNumber(green + "_min"),
            },
        ).rename("plume_green"),
        None,
        True,
    )

    image = image.addBands(
        image.expression(
            f"(b('{red}') > MINI && b('{red}') < MAXI) ? 1 : 0",
            {
                "MAXI": limits.getNumber(red + "_max"),
                "MINI": limits.getNumber(red + "_min"),
            },
        ).rename("plume_red"),
        None,
        True,
    )

    image = image.addBands(
        image.expression(
            "(b('plume_blue') + b('plume_green') + b('plume_red'))"
        ).rename("plume"),
        None,
        True,
    )

    plume_mask = image.select("plume").gt(0.5)

    pixel_count = plume_mask.select(0).connectedPixelCount(100, False)
    count_mask = pixel_count.select(0).gt(50)

    image = image.updateMask(plume_mask).updateMask(count_mask)

    return image



"""Submodule for genarate statistical plots"""
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from typing import Union
from matplotlib.axes import Axes
from matplotlib.ticker import NullLocator
from numpy.typing import ArrayLike
from .stats import corr_matrix


def plot_corr_matrix(
    data: pd.DataFrame,
    variables: Union[ArrayLike, None] = None,
    half: bool = False,
    hide_insignificants: bool = False,
    significant_threshold: float = 0.05,
    show_labels: bool = True,
    show_colorbar: bool = False,
    palette: str = "Spectral",
    text_color: str = "black",
    ax: Union[Axes, None] = None,
) -> Axes:
    """
    Calculate the pearson correlation matrix of the variables in a dataframe.

    Parameters
    ----------
    data : pd.DataFrame
        Dataframe with the variables to evaluate their correlation.

    variables : ArrayLike | None = None
        The variables of interest, if it is not defined, all variables in
        the dataframe will be evaluated.

    half : bool (optional)
        If True, only show the corerlation of the first half of the matrix,
        excluding the repeated correlation.

    hide_insignifcants : bool (optional)
        If True, hide all the correlation with a p-value greater than the
        significant threshold.

    significant_threshold : float (optional)
        Threshold of significant correlation.

    show_labels : bool (optional)
        Show the correlation value.

    show_colorbar : bool (optional)
        Show colorbar

    palette : str (optional)
        Color palette for correlation plot.

    text_color : str (optional)
        Color of text correlation labels.

    ax : matplotlib.axes.Axes | None (optional)
        Axes to draw the correlation matrix.
    returns
    -------
    ax : matplotlib.axes.Axes
        Correltion matrix.
    """
    # If variables are not defined get all columns from data
    if variables is None:
        variables = data.columns

    # Get the number of variables
    N = len(variables)

    # Reverse variables for plot
    reverse = variables[::-1]

    # Get the correlation matrix
    corr = corr_matrix(
        data, variables, half, hide_insignificants, significant_threshold
    )

    # If there not axes create one
    if ax is None:
        fig = plt.figure()
        ax = fig.add_subplot(1, 1, 1)

    # Plot matrix with pcolormesh
    im = ax.pcolormesh(
        variables, reverse, corr, cmap=palette, edgecolor="w", vmin=-1, vmax=1
    )

    # Invert y axis
    ax.invert_yaxis()

    # Add the colorbar
    if show_colorbar:
        cax = ax.inset_axes([1.04, 0.1, 0.05, 0.8])
        bar = plt.colorbar(im, cax=cax, label="Correlation")

    if show_labels:
        x, y = np.meshgrid(np.arange(N), np.arange(N))
        x = x.reshape(-1)
        y = y.reshape(-1)
        t = corr.values.reshape(-1)

        for xi, yi, ti in zip(x, y, t):
            if np.isfinite(ti):
                ax.text(
                    xi, yi,
                    round(ti, 2),
                    color=text_color,
                    size=8,
                    ha="center",
                    va="center",
                )

    ax.xaxis.set_minor_locator(NullLocator())
    ax.yaxis.set_minor_locator(NullLocator())

    return ax


"""
Function to process Landsat SR Images.
"""
import ee


def scaler(image: ee.Image) -> ee.Image:
    """
    Scale optical bands to surface reflectance values and thermal bands to Kelvin.

    Parameters
    ----------
    image : ee.Image
        Landsat SR image.

    Returns
    -------
    image : ee.Image
        Image with bands scaled.
    """
    optical = image.select("SR_B.").multiply(0.0000275).add(-0.2)
    thermal = image.select("ST_B.*").multiply(0.00341802).add(149)

    image = image.addBands(optical, None, True).addBands(thermal, None, True)

    return image


def cloud_mask(image: ee.Image, mask_snow:bool = False) -> ee.Image:
    """
    Mask pixels classifed as clouds from QA_PIXEL band.

    Parameters
    ----------
    image : ee.Image
        Image to mask.

    mask_snow : bool (optional)
        If `True` mask pixels classified as snow.
        If `False` mask only pixels classified as cloud.

    Returns
    -------
    image : ee.Image
        Masked image.
    """
    qa = image.select("QA_PIXEL")

    cirrus = qa.bitwiseAnd((1 << 2)).eq(0)
    cloud = qa.bitwiseAnd((1 << 3)).eq(0)
    shadow = qa.bitwiseAnd((1 << 4)).eq(0)
    snow = qa.bitwiseAnd((1 << 5)).eq(0)

    image = image.updateMask(cirrus).updateMask(cloud).updateMask(shadow)

    if mask_snow:
        image = image.updateMask(snow)

    return image



"""Sample pixel values from image in a region of interest"""
import ee
import numpy as np
from numpy.typing import ArrayLike


ee.Initialize()


def sample_image(
    image: ee.Image,
    geom: ee.Geometry,
    scale: float,
    band: str,
) -> ArrayLike:
    """
    Sample all pixels from an image in the specified band.

    Parameters
    ----------
    image : ee.Image
        Image to sample.

    geom : ee.Geometry
        Region of interest.

    scale : float
        Scale of image to sample.

    band : str
        Band of interest.

    Returns
    -------
    """
    image = image.select(band)

    try:
        sample = image.sampleRegions(collection=geom, scale=scale, geometries=False)

        sample = sample.toList(sample.size())
        raw = sample.getInfo()

        data = np.array([r["properties"][band] for r in raw])
    except:
        data = np.array([np.nan])

    return data


def sample_collection(
    image_collection: ee.ImageCollection,
    geom: ee.Geometry,
    scale: float,
    band: str,
) -> list:
    """
    This function sample all images in an image collection applying the sample_image
    function to all images.

    Parameters
    ----------
    image_collection : ee.ImageCollection
        Image collection to sample.

    geom : ee.Geometry
        Region of interest.

    scale : float
        Scale of image to sample.

    band : str
        Band of interest.

    Returns
    -------
    data : list
        list of np.array with all the sampled values per image.
    """
    N = image_collection.size().getInfo()
    ic_list = image_collection.toList(N)

    data = []

    for i in range(N):
        image = ee.Image(ic_list.get(i))
        values = sample_image(image, band=band, geom=geom, scale=scale)
        data.append(values)

    return data



"""Functions to process image collections as time series"""
import ee
import pandas as pd


ee.Initialize()


def extract_dates(image_collection: ee.ImageCollection) -> pd.Series:
    """
    Extract dates from all images in an image collection.

    Parameters
    ----------
    image_collection : ee.ImageCollection
        Colletion of image.

    Returns
    -------
    dates : pd.Series
        Series with the dates of the images.
    """
    dates = (
        image_collection.reduceColumns(ee.Reducer.toList(), ["system:time_start"])
                        .get("list")
                        .getInfo()
    )

    dates = pd.DatetimeIndex(pd.to_datetime(dates, unit="ms"))

    return dates


def trend(image_collection: ee.ImageCollection, band: str) -> ee.ImageCollection:
    """
    Calculate the linear trend and stational variation to an image collection.

    Parameters
    ----------
    image_collection : ee.imageCollection
        Image collection to perform the linear trend and detrend.

    band : str
        Name of the band to perform the calc.

    Returns
    -------
    image_collection : ee.ImageCollection
        Image collection with the Raw data, Time, Trend and Stational variation bands.
    """
    def time_func(image: ee.Image) -> ee.Image:
        """Calculate time band for linear regression"""
        time = (
            image.metadata("system:time_start")
                 .divide(1000*60*60*24*365)
                 .rename("time")
        )

        return image.addBands(time)

    mean = image_collection.select(band).mean()

    image_collection = image_collection.map(time_func)
    image_collection = image_collection.select(["time", band])

    fitted = image_collection.reduce(ee.Reducer.linearFit())

    def pred_func(image: ee.Image) -> ee.Image:
        """Calculate linear regression"""
        pred = (
            image.select("time")
                 .multiply(fitted.select("scale"))
                 .add(fitted.select("offset"))
                 .rename("predicted")
                 .toFloat()
        )

        return image.addBands(pred)

    def stat_func(image: ee.Image) -> ee.Image:
        """Calculate the stational component"""
        stat = image.expression(
            "band - pred + mean",
            {
                "band":image.select(band),
                "pred":image.select("predicted"),
                "mean":mean,
            },
        ).rename("detrended").toFloat()

        return image.addBands(stat)

    image_collection = image_collection.map(pred_func).map(stat_func)

    return image_collection


def reduce_by_year(
    image_collection: ee.ImageCollection,
    reducer: ee.Reducer,
    start: int,
    end: int,
) -> ee.ImageCollection:
    """
    Resample image collection to annual reduced collection.

    Parameters
    ----------
    image_collection : ee.ImageCollection
        Image collection to reduce.

    reducer : ee.Reducer
        Reducer to apply throught the image collection.

    start : int
        First year.

    end : int
        Last year.

    Returns
    -------
    annual_collection : ee.ImageCollection
        Image collection reduced to annual frequency.
    """
    annual_collection = []

    for year in range(start, end+1):
        img = (
            image_collection.filter(ee.Filter.calendarRange(year, year, "year"))
                            .reduce(reducer)
                            .set("year", year)
        )

        annual_collection.append(img)

    return ee.ImageCollection(annual_collection)


def reduce_by_month(
    image_collection: ee.ImageCollection,
    reducer: ee.Reducer,
) -> ee.ImageCollection:
    """
    Reduce image collection to monthly statistic.

    Parameters
    ----------
    image_collection : ee.ImageCollection
        Image collection to reduce.

    reducer : ee.Reducer
        Reducer to apply throught image collection.

    Returns
    -------
    monthly_collection : ee.ImageCollection
        Image collection reduced to monthly frequency.
    """
    monthly_collection = []

    for month in range(1, 13):
        img = (
            image_collection.filter(ee.Filter.calendarRange(month, month, "month"))
                            .reduce(reducer)
                            .rename("seasonal")
                            .set("month", month)
        )

        monthly_collection.append(img)

    return ee.ImageCollection(monthly_collection)


def calculate_anomalies(
    image_collection: ee.ImageCollection, monthly_mean: ee.ImageCollection
) -> ee.ImageCollection:
    """
    Calculate the anomalies of a ImageCollection subtracting the monthly mean values.

    Parameters
    ----------
    image_collection : ee.ImageCollection
        Image collection trended by the trend function to calc the anomalies.

    monthly_mean : ee.ImageCollection
        Image collection with the monthly means of ImageCollection calculated by the
        `reduce_by_month` function.

    Returns
    -------
    anomalies : ee.ImageCollection
        ImageCollection with stational means added and anomalies calcualted.
    """
    def calc_anomaly(image: ee.Image) -> ee.Image:
        """Function for calculate anomalies."""
        anomaly = image.expression(
            "stat - mean",
            {"stat": image.select("detrended"), "mean": image.select("seasonal")},
        ).rename("anomaly")

        return image.addBands(anomaly)

    m01 = image_collection.filter(ee.Filter.calendarRange(1, 1, "month")).map(
        lambda x: x.addBands(monthly_mean.filter(ee.Filter.eq("month", 1)).first())
    )
    m02 = image_collection.filter(ee.Filter.calendarRange(2, 2, "month")).map(
        lambda x: x.addBands(monthly_mean.filter(ee.Filter.eq("month", 2)).first())
    )
    m03 = image_collection.filter(ee.Filter.calendarRange(3, 3, "month")).map(
        lambda x: x.addBands(monthly_mean.filter(ee.Filter.eq("month", 3)).first())
    )
    m04 = image_collection.filter(ee.Filter.calendarRange(4, 4, "month")).map(
        lambda x: x.addBands(monthly_mean.filter(ee.Filter.eq("month", 4)).first())
    )
    m05 = image_collection.filter(ee.Filter.calendarRange(5, 5, "month")).map(
        lambda x: x.addBands(monthly_mean.filter(ee.Filter.eq("month", 5)).first())
    )
    m06 = image_collection.filter(ee.Filter.calendarRange(6, 6, "month")).map(
        lambda x: x.addBands(monthly_mean.filter(ee.Filter.eq("month", 6)).first())
    )
    m07 = image_collection.filter(ee.Filter.calendarRange(7, 7, "month")).map(
        lambda x: x.addBands(monthly_mean.filter(ee.Filter.eq("month", 7)).first())
    )
    m08 = image_collection.filter(ee.Filter.calendarRange(8, 8, "month")).map(
        lambda x: x.addBands(monthly_mean.filter(ee.Filter.eq("month", 8)).first())
    )
    m09 = image_collection.filter(ee.Filter.calendarRange(9, 9, "month")).map(
        lambda x: x.addBands(monthly_mean.filter(ee.Filter.eq("month", 9)).first())
    )
    m10 = image_collection.filter(ee.Filter.calendarRange(10, 10, "month")).map(
        lambda x: x.addBands(monthly_mean.filter(ee.Filter.eq("month", 10)).first())
    )
    m11 = image_collection.filter(ee.Filter.calendarRange(11, 11, "month")).map(
        lambda x: x.addBands(monthly_mean.filter(ee.Filter.eq("month", 11)).first())
    )
    m12 = image_collection.filter(ee.Filter.calendarRange(12, 12, "month")).map(
        lambda x: x.addBands(monthly_mean.filter(ee.Filter.eq("month", 12)).first())
    )

    anomalies = (
        m01.merge(m02)
           .merge(m03)
           .merge(m04)
           .merge(m05)
           .merge(m06)
           .merge(m07)
           .merge(m08)
           .merge(m09)
           .merge(m10)
           .merge(m11)
           .merge(m12)
    )

    anomalies = anomalies.sort("system:time_start")
    anomalies = anomalies.map(calc_anomaly)

    return anomalies


def time_series_processing(
    image_collection: ee.ImageCollection, band: str
) -> list:
    """
    This function take an ee.ImageCollection and calculate the linear trend, the
    detrended component, the seasonal mean and anomalies for the selected band.
    To calculate the linear trend the function use `ee.Reducer.linearFit()`, next,
    calculate the detrended component subtracting the linear_fitted values to the
    original series and restoring its mean, the monthly means are calculeted by
    selecting all the image in the specific month and reduce it by its mean, finally,
    subtracting the monthly means to the detrended component anomalies are obtained.
    This function work as a "wrapper" function of `trend()`, `reduce_by_month()` and
    `calc_anomalies()`.

    Parameters
    ----------
    image_collection : ee.ImageCollection
        Image collection to proces.

    band : str
        Name of the band of interest.

    Returns
    -------
    data : ee.ImageCollection
        Image collection with the raw data, the linear trend, deternded component,
        seasonal mean and the anomalies.

    monthly_mean : ee.ImageCollection
        Image collection with the twelves monthly means calculated.
    """
    trended = trend(image_collection, band)
    monthly_mean = reduce_by_month(trended.select("detrended"), ee.Reducer.mean())
    data = calculate_anomalies(trended, monthly_mean)

    return data, monthly_mean



"""Detect changes in coastlines"""
import ee


ee.Initialize()


def shoreline_characterization(
    image: ee.Image,
    region: ee.Geometry,
    scale: float,
    band: str,
) -> ee.FeatureCollection:
    """
    Extract shoreline from image classified.

    Parameters
    ----------
    image : ee.Image
        Classified image to analyse.

    region : ee.Geometry
        Region of interest to clip the image and shorelines.

    scale : float
        Image scale.

    band : str
        Band of interest.

    Returns
    -------
    polylines : ee.FeatureCollection
        Digitalized shorelines from images.
    """
    def polygon_to_polyline(geom):
        """convert polygon to polyline"""
        coords = geom.geometry().simplify(maxError=scale).coordinates()
        buffer = ee.Number(scale).multiply(-1)

        polyline = geom.setGeometry(
            ee.Geometry.MultiLineString(coords)
                       .intersection(region.buffer(buffer))
        )

        return polyline

    image = image.select(band)

    polygon = image.clip(region).reduceToVectors(scale=scale)       # Polygonize image
    polygon = polygon.filter(ee.Filter.eq("label", 0))              # Select land

    polylines = polygon.map(polygon_to_polyline)

    return polylines



"""Submodule to ease some statisticals"""
import numpy as np
import pandas as pd
from typing import Union
from scipy.stats import pearsonr
from numpy.typing import ArrayLike


def corr_matrix(
    data: pd.DataFrame,
    variables: Union[ArrayLike, None] = None,
    half: bool = False,
    hide_insignificants: bool = False,
    singificant_threshold: float = 0.05,
) -> pd.DataFrame:
    """
    Calculate the pearson correlation matrix of the variables in a dataframe.

    Parameters
    ----------
    data : pd.DataFrame
        Dataframe with the variables to evaluate their correlation.

    variables : ArrayLike | None (optional)
        The variables of interest, if it is not defined, all variables in
        the dataframe will be evaluated.

    half : bool (optional)
        If True, only show the corerlation of the first half of the matrix,
        excluding the repeated correlation.

    hide_insignifcants : bool (optional)
        If True, hide all the correlation with a p-value greater than the
        significant threshold.

    siginificant_threshold : float (optional)
        Threshold of significant correlation.

    returns
    -------
    corr : pd.DataFrame
        Dataframe with the correlation values.
    """
    if variables is None:
        variables = data.columns

    reverse = variables[::-1]

    N = len(variables)

    corr = np.empty((N, N))
    pval = np.full((N, N), np.nan)
    mask = np.full((N, N), np.nan)

    for i, iv in enumerate(variables):
        for j, jv in enumerate(reverse):
            c, p = pearsonr(data[iv], data[jv])
            corr[j, i] = c

            if p <= singificant_threshold:
                pval[j, i] = 1.0

        mask[: N - i, i] = 1.0

    if half:
        corr *= mask

    if hide_insignificants:
        corr *= pval

    corr = pd.DataFrame(data=corr, index=reverse, columns=variables)

    return corr



"""Get hypsometric curves from a DEM"""
import ee
import numpy as np
from numpy.typing import ArrayLike
from .sample import sample_image


ee.Initialize()


def hypsometric_curve(
    catchment: ee.Geometry,
    dem: ee.Image,
    scale: float,
    band: str,
    bins: int = 100,
) -> list:
    """
    Calculate the cumulative area and comulative height from a catchment for the
    hypsometric curve.

    Parameters
    ----------
    catchment : ee.Geometry
        Catchment area to analyse.

    dem : ee.Image
        Digital elevation model (DEM) to analyse.

    scale : float
        Scale of DEM.

    band : str
        Band of DEM with elevetaion information.

    bins : int
        Number of bins for cumalitive area an height.

    Returns
    -------
    area : ArrayLike
        Normalized area.

    height : ArrayLike
        Normalized height.
    """
    dem = dem.select(band)

    values = sample_image(dem, catchment, scale, band)
    values = values/values.max()

    area, height = np.histogram(values, bins=np.linspace(0, 1, bins+1))

    area = 1 - np.flip(area.cumsum()/area.cumsum().max())
    height = np.flip(height[:-1])

    return area, height



"""Classify pixels in images"""
import ee

ee.Initialize()


def water_coverage(
    image: ee.Image,
    blue: str,
    green: str,
    red: str,
    nir: str,
    swir: str,
) -> ee.Image:
    """
    Function to classify pixels as water based on Based on Zou et al. (2017) and
    Worder & Beurs (2020).

    Parameters
    ----------
    image : ee.Image
        Image to classify.

    blue : str
        Blue band name.

    green : str
        Green band name.

    red : str
        Red band name.

    nir : str
        Near infrared band name.

    swir : str
        Short-wave infrared band name.

    Returns
    -------
    image : ee.Image
        Image with water coverage band classified where water pixels are 1 and
        no-water are 0.
    """
    mndwi = image.expression(
        f"(b('{green}') - b('{swir}'))/(b('{green}') + b('{swir}'))"
    ).rename("mNDWI")
    ndvi = image.expression(
        f"(b('{nir}') - b('{red}'))/(b('{nir}') + b('{red}'))"
    ).rename("NDVI")
    evi = image.expression(
        f"2.5*(b('{nir}') - b('{red}'))/(b('{nir}') + 6*b('{red}') - 7.5*b('{blue}') + 1)"
    ).rename("EVI")

    water = (
        evi.lt(0.1)
           .And(mndwi.gt(evi).Or(mndwi.gt(ndvi)))
           .rename("WATER_COVERAGE")
    )

    image = image.addBands(water)

    return image


def vegetation_coverage(
    image: ee.Image,
    blue: str,
    red: str,
    nir: str,
    swir: str,
) -> ee.Image:
    """
    Function to classify pixels as vegetation based on Based on Zou et al. (2017)
    and Worder & Beurs (2020).

    Parameters
    ----------
    image : ee.Image
        Image to classify.

    blue : str
        Blue band name.

    green : str
        Green band name.

    red : str
        Red band name.

    nir : str
        Near infrared band name.

    swir : str
        Short-wave infrared band name.

    Returns
    -------
    image : ee.Image
        Image with vegettaion coverage band classified where vegetation pixels
        are 1 and no-vegetation are 0.
    """
    ndbi = image.expression(
        f"(b('{swir}') - b('{nir}'))/(b('{swir}') + b('{nir}'))"
    ).rename("NDBI")
    ndvi = image.expression(
        f"(b('{nir}') - b('{red}'))/(b('{nir}') + b('{red}'))"
    ).rename("NDVI")
    evi = image.expression(
        f"2.5*(b('{nir}') - b('{red}'))/(b('{nir}') + 6*b('{red}') - 7.5*b('{blue}') + 1)"
    ).rename("EVI")

    vegetation = (
        evi.gte(0.1)
           .And(ndvi.gte(0.2).And(ndbi.lt(0.0)))
           .rename("VEGETATION_COVERAGE")
    )

    image = image.addBands(vegetation)

    return image


def clean_boolean_classification(
    image: ee.Image,
    scale: float,
    band: str,
    true_class_threshold: float,
    false_class_treshold: float,
) -> ee.Image:
    """
    Clean boolean coverage classification counting the neighbors pixels and filter
    those which surpass their respective thresholds.

    Parameters
    ----------
    image : ee.Image
        Boolean classification raster to clean.

    scale : float
        Image scale to perform the process.

    band : str
        Band name with the classification.

    true_class_threshold : float
        Threshold for true values to conserve an entity.

    false_class_treshold : float
        Threshold for false values to conserve an entity.

    Returns
    -------
    filled : ee.Image
        Boolean classification image cleaned.
    """
    image = image.select(band).toInt()

    connected_false = ee.Number(true_class_threshold).divide(scale).int()
    connected_true = ee.Number(false_class_treshold).divide(scale).int()

    false_filled = (
        image.addBands(image)
        .reduceConnectedComponents(ee.Reducer.median(), band, connected_false)
        .unmask(99)
        .eq(99)
        .And(image.neq(0))
    )

    filled = (
        false_filled.addBands(false_filled)
        .reduceConnectedComponents(ee.Reducer.median(), band+"_1", connected_true)
        .unmask(99)
        .eq(99)
        .And(false_filled.neq(1))
    )

    filled = filled.expression("b(0) == 0 ? 1 : 0").rename(band)

    return filled



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['statgis', 'statgis.gee', 'statgis.statutils']

package_data = \
{'': ['*']}

install_requires = \
['earthengine-api>=0.1.335',
 'matplotlib>=3.2.2',
 'pandas>=1.3.5',
 'scipy>=1.7.3']

setup_kwargs = {
    'name': 'statgis',
    'version': '1.0.5',
    'description': 'Tools for improve work with geospatial data in Python',
    'long_description': '# StatGIS Python Package\n\n![PyPI](https://img.shields.io/pypi/v/statgis) ![PyPI - Python Version](https://img.shields.io/pypi/pyversions/statgis?color=yellow) [![Documentation Status](https://readthedocs.org/projects/statgis/badge/?version=latest)](https://statgis.readthedocs.io/en/latest/?badge=latest)\n\nStatGIS help to the users to process and analyse satellite images in Google Earth Engine.\n\n## Installation\n\n```bash\n$ pip install statgis\n```\n\n## Usage\n\nIn this package you could find function to scale optical images, classify images,\ncharacterize shorelines and much more.\n\n```python\n# Example: mask clouds and scale to reflectance values\nimport ee\nfrom statgis.gee import landsat_functions\n\nee.Initialize()\n\nimage_collection = (\n    ee.ImageCollection("LANDSAT/LC09/C02/T1_L2")\n      .map(landsat_functions.cloud_mask)\n      .map(landsat_functions.scaler)\n)\n```\n\n## Contributing\n\nInterested in contributing? Check out the contributing guidelines. Please note that this project is released with a Code of Conduct. By contributing to this project, you agree to abide by its terms.\n\n## License\n\n`statgis` was created by Sebástian Narváez-Salcedo, Brayan Navarro-Londoño. It is licensed under the terms of the MIT license.\n\n## Credits\n\n`statgis` was created with [`cookiecutter`](https://cookiecutter.readthedocs.io/en/latest/) and the `py-pkgs-cookiecutter` [template](https://github.com/py-pkgs/py-pkgs-cookiecutter).\n',
    'author': 'Sebástian Narváez-Salcedo',
    'author_email': 'None',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.12',
}


setup(**setup_kwargs)



"""
Function to process Sentinel-2 images.
"""
import ee

def scaler(image: ee.Image) -> ee.Image:
    """
    Scale bands from Sentinel-2 image.

    Parameters
    ----------
    image : ee.Image
        Sentinel-2 image to scale.

    Returns
    -------
    image : ee.Image
        Image with bands scaled.
    """
    bands = image.select("B.*").divide(10000)
    image = image.addBands(bands, None, True)

    return image


def cloud_mask(image: ee.Image) -> ee.Image:
    """
    Mask clouds from Sentinel-2 image using QA60 band.

    Parameters
    ----------
    Image : ee.Image
        Image to mask.

    Returns
    -------
    Image : ee.Image
        Masked image.
    """
    qa = image.select("QA60")

    clouds = qa.bitwiseAnd((1 << 10)).eq(0)
    cirrus = qa.bitwiseAnd((1 << 11)).eq(0)

    image = image.updateMask(clouds).updateMask(cirrus)

    return image
